#!/opt/conda/bin/python
import sys, getopt
import requests
import json
import sys
import os
import time
import math
import os.path as path
# Import getopt module
import getopt
import getpass
from requests.exceptions import ConnectionError
from datetime import datetime, timedelta

MAAP_ENV_TYPE = os.getenv("MAAP_ENV_TYPE")
CLIENT_ID = os.getenv("CLIENT_ID")
HOME = os.getenv("HOME")

#Url to retrieve the access token with the user credential
AUTH_ACCESS_TOKEN_URL = os.getenv("AUTH_ACCESS_TOKEN_URL") 
BEARER=""
#if windows we take the current folder
if sys.platform == 'win32':
   USER_INFO_FILE_PATH=os.getcwd()+"\maap-s3-userinfo.json"
   USER_LAST_UPLOAD_INFO_FILE_PATH=os.getcwd()+"\maap-s3-multipartinfo.json"
else :
   USER_INFO_FILE_PATH=HOME+"/.maap/maap-s3-userinfo.json"
   USER_LAST_UPLOAD_INFO_FILE_PATH=HOME+"/.maap/maap-s3-multipartinfo.json"
   
userinfo = {}
multipartinfo = {}

def display_help():
    print('Usage: [option...] {upload|download|list|delete|refresh|resume}')
    # print('-i                                                                   Get a fresh token before any
    # request. It ask for email and password')
    print('\033[1m' + 'Options:' + '\033[0m')
    print('\t-p, --private\n' + 2*'\t' + 'Private data, uses a specific bucket for users. Include this option in your'
                                         ' command to download, upload and delete files from your user data storage.')
    print('\033[1m' + 'Commands:' + '\033[0m')
    print('\tupload     myFile.tiff locally          path/myFile.tiff in the S3    Upload data in the S3')
    print('\tdownload   myFileName.tiff              path/in/S3/file.tiff          Download a data from the S3')
    print('\tlist       folder/path                                                List paths that start with the input')
    print('\tdelete     path/in/S3/file.tiff                                       Delete an existing data on S3')
    print('\trefresh                                                               Refresh credentials and password')
    print('\ttoken      email                        password                      Return a bearer token')
    print('\tlogin      email                        password                      Login the user')
    print('\tresume' + 8*'\t' + '      Resume last interrupted multipart upload')
    sys.exit(2)


#########################
# Init the bearer       #
#########################
def init():
    if os.path.isfile(USER_INFO_FILE_PATH):
        print("[INFO] Personal user info is found")
    else:
        print("[INFO] Personal user info is not found")
        refresh()


###########################
# Regenerate token using user email and password #
###########################
def refresh():
    email = input("Your email: ")
    #password
    password = getpass.getpass('Your password:')
    #Function to generate a new token
    generate_token(email, password)

###################################
# Login using email dans password #
###################################
def login(email, password):
    if email and password:
        print("[INFO] Get an existing or fresh token")         
        #Function to generate a new token
        generate_token(email, password)
    else:
        print("[ERROR] Please check your email or password") 

###########################
# Generate token and save #
###########################
def generate_token(email, password): 
        
    print("[INFO] Start retrieving token for authent")
    #Set the bearer
    url = AUTH_ACCESS_TOKEN_URL
    print (url)
    print (CLIENT_ID)
    response = requests.post(url, data={'client_id': CLIENT_ID, 'username': email, 'password': password, "grant_type": "password", "scope": "profile"})
    if response.status_code == 401:
        print("Authentication failed")
        sys.exit(1)
    print(response)
    print(response.text)
    #Convert the string to json to fecth access_token
    data = json.loads(response.text)
    token = data['access_token']
    token_expires_in = data['expires_in']
    refresh_token = data['refresh_token']
    refresh_expires_in = data['refresh_expires_in']

    # add the token in the json info file
    #Create a json with email and password
    userinfo = {
        'email': email,
        'token': token,
        'token_expires_in': token_expires_in,
        'refresh_token': refresh_token,
        'refresh_expires_in': refresh_expires_in,
        'auth_date': datetime.now().isoformat()
    }

    if token: 
        #add the json in the file
        with open(USER_INFO_FILE_PATH, 'w') as outfile:
            json.dump(userinfo, outfile, default=str)
           
        print("[INFO] Token saved for one hour and ready to be used "+token)
        
    else:
        print("[ERROR] Token is empty. Please 1) run refresh (-r) function and check your password")
        # Terminate the script
        sys.exit(2)


################################
# Generate token and return it #
###############################
def get_token(email, password): 
        
    #print("[INFO] Start retrieving token for authent")
    #Set the bearer
    url = AUTH_ACCESS_TOKEN_URL
    response = requests.post(url, data={'client_id': CLIENT_ID, 'username': email, 'password': password, "grant_type": "password", "scope": "profile"})
    #Convert the string to json to fecth access_token
    data = json.loads(response.text)
    token = data['access_token']
    print (token)
    return token


#########################
# Check if file is older#
#########################
def is_file_older_than_x_hour(file, hour=1): 
    file_time = path.getmtime(file) 
    # Check against 214 hour 
    return ((time.time() - file_time) > 3600*hour)



##### refresh token
def _refresh_token(email, refresh_token):
    '''refresh existing token for a new one'''
    url = AUTH_ACCESS_TOKEN_URL
    data = {'grant_type': 'refresh_token',
            'refresh_token': refresh_token,
            'client_id': CLIENT_ID
            }
    response = requests.post(url, data=data)
    if response.status_code != 200:
        print(response , response.text)

    auth_data = response.json()

    userinfo = {
        'email': email,
        'token': auth_data.get('access_token'),
        'token_expires_in': auth_data.get('expires_in'),
        'refresh_token': auth_data.get('refresh_token'),
        'refresh_expires_in': auth_data.get('refresh_expires_in'),
        'auth_date': datetime.now().isoformat()
    }

    if auth_data.get('access_token'):
        # don't store creds in plaintext in a real app obviously
        with open(USER_INFO_FILE_PATH, 'w') as outfile:
            json.dump(userinfo, outfile)
    else:
        print('There was an error refreshing your access token')
        print(r.text)


def verify_token(func):

    def inner(*args, **kwargs):
        init()
        with open(USER_INFO_FILE_PATH) as json_file:
            userinfo = json.load(json_file)
        for item in ['token', 'refresh_token','auth_date', 'token_expires_in', 'refresh_expires_in']:
            if userinfo.get(item) is None:
                refresh()
        auth_date = userinfo.get('auth_date')
        if datetime.now() > datetime.fromisoformat(auth_date) + timedelta(0,userinfo.get('token_expires_in')):
            print('token expired')
            if datetime.now() > datetime.fromisoformat(auth_date) + timedelta(0,userinfo.get('refresh_expires_in')):
                refresh()
            else:
                print('refreshing token...')
                _refresh_token(email=userinfo.get('email'), refresh_token=userinfo.get('refresh_token'))
        return func(*args, **kwargs)
    return inner


#########################
# Upload the data in S3 #
#########################
@verify_token
def upload(sourceFile, destination, private: bool):
    print("[INFO] Source file is : ", sourceFile)
    print("[INFO] Destination file is : ", destination) 

    if sourceFile and destination:
        print("[INFO] Get an existing or fresh token")

        # If the file is less that 100 MB we upload directly
        #Check file size
        fileSize = os.stat(sourceFile).st_size
        print("Size "+ str(fileSize))
    
        #We have more than 5GB
        if fileSize > 5000000000:
            #We upload the multi part data
            print("[INFO] Starting multi part upload")
            upload_multipart(sourceFile, destination, private)

        else: 
            with open(USER_INFO_FILE_PATH) as json_file:
                userinfo = json.load(json_file)
                #Get the info
                token=userinfo['token']

            print("[INFO] Starting retrieving the presigned url for the creation of the file with token "+ token)
            #files = {'upload_file': open(sourceFile,'rb')}
            url = "https://gravitee-gateway."+MAAP_ENV_TYPE.lower()+".esa-maap.org/s3/"+destination

            response = requests.put(url, headers={'Authorization': 'Bearer ' + token},
                                    params={'private': private}, allow_redirects=False)
            if response.status_code == 401 or response.status_code == 403:
                print(response.content)
                sys.exit(1)
            location = response.headers['Location']
            print("[INFO] Location is "+ location)

            if location:
                print("[INFO] Start uploading the file")
                with open(sourceFile, 'rb') as f:
                    response = requests.put(location, data=f)
                    print(response)
                #files = {'file': open(sourceFile, 'rb')}
                #r = requests.put(location, files=files)
                
            else:
                print("[ERROR] Presigned url not generated. Please re run refresh or contact admin if the error persist")

    else:
        display_help()


###########################################################
# Upload the data in S3, the data is split chunk by chunk #
###########################################################
@verify_token
def upload_multipart(sourceFile, destination, private_bucket: bool):
    # Get the token
    with open(USER_INFO_FILE_PATH) as json_file:
        userinfo = json.load(json_file)
        #Get the info
        token=userinfo['token']

    #Set variables
    filePath = sourceFile
    if "/" in destination and not private_bucket:
        key = destination.split('/', 1)[1]
        bucketName = destination.split('/')[0]
    else:
        key = destination
        bucketName = ""

    fileSize = os.stat(filePath).st_size
    print("Size "+ str(fileSize))
    #Set max to split to 5M
    max_size = 5 * 1024 * 1024 # Approach 1: Assign the size
    nbParts = math.ceil(fileSize/max_size)    #calculate nbParts
    print("[INFO] We will have "+ str(nbParts)+" parts")

    url = "https://gravitee-gateway."+MAAP_ENV_TYPE.lower()+".esa-maap.org/s3/generateUploadId"
    params={'bucketName': bucketName, 'objectKey': key, 'private': private_bucket}
    response = requests.get(url, params=params, headers={'Authorization': 'Bearer ' + token})
    if response.status_code == 401 or response.status_code == 403:
        print(response.content)
        sys.exit(1)

    print("[INFO] uploadId "+ response.text)
    #Save upload id
    uploadId = response.text

    #Generate presigned urls 
    url = "https://gravitee-gateway."+MAAP_ENV_TYPE.lower()+".esa-maap.org/s3/generateListPresignedUrls"
    params = {'bucketName': bucketName, 'objectKey': key, 'nbParts': nbParts, 'uploadId': uploadId,
              'private': private_bucket}
    response = requests.get(url, params = params, headers = {'Authorization': 'Bearer '+token})
    if response.status_code == 401 or response.status_code == 403:
        print(response.content)
        sys.exit(1)

    stringList = response.text
    str1 = stringList.replace(']','').replace('[','')
    listPresignedUrl  = str1.replace('"','').split(",")

    # we load the data
    #print(listPresignedUrl)
    parts = []

    #sys.stdout = open("log.txt", "w")
    with open(filePath, 'rb') as f:    
        i = 0
        while i < nbParts:
            print("Upload part "+ str(i+1))
            file_data = f.read(max_size)
            headers={'Content-Length': str(max_size)}
            #print(listPresignedUrl[i])
            try:
                response = requests.put(listPresignedUrl[i], data=file_data, headers=headers)
            except ConnectionError as e:
                print("Error 401 ConnectionError")
                print("Retry...")
                time.sleep(60)
                resume()

            if response.status_code == 401 or response.status_code == 403:
                print(response.content)
                sys.exit(1)
            #print(response.headers)
            #print(response.text)
            etag = response.headers['ETag']  
            parts.append({'eTag': etag, 'partNumber': int(i+1)})
            print(parts)
            i = i+1
            #We save also the multipart
            #So we can resume it if upload failed
            multipartinfo = {
                'uploadId': uploadId,
                'partsUpploaded': parts,
                'sourceFile': sourceFile,
                'destination': destination,
                'isPrivate': private_bucket
            }
            
            #add the json in the file
            with open(USER_LAST_UPLOAD_INFO_FILE_PATH, 'w') as outfile:
                json.dump(multipartinfo, outfile)


    #sys.stdout.close()
    print("Complete multipart upload")
    #complete the multi part
    url = "https://gravitee-gateway."+MAAP_ENV_TYPE.lower()+".esa-maap.org/s3/completeMultiPartUploadRequest"
    params = {'bucketName': bucketName, 'objectKey': key, 'nbParts': nbParts, 'uploadId': str(uploadId),
              'private': private_bucket}
    response = requests.get(url, data=str(parts), params=params, headers={'Authorization': 'Bearer ' + token})
    if response.status_code != 200:
        print(response.content)
        sys.exit(1)
    #delete the file of multipart info because upload was success
    print("Upload completed")
    os.remove(USER_LAST_UPLOAD_INFO_FILE_PATH) 


###################################
# Resume failed multi part upload #
###################################
@verify_token
def resume():
    print("[INFO] Resume the last multipart upload")
    print("[INFO] Check last multipart upload metadata")

    if os.path.isfile(USER_LAST_UPLOAD_INFO_FILE_PATH):
        
        #Generate or get a token
        print("[INFO] Previous multi part upload file found")
        token=''
        #Get the token
        with open(USER_INFO_FILE_PATH) as json_file:
            userinfo = json.load(json_file)
            #Get the info
            token=userinfo['token']
        
        #Get the data in the json file
        with open(USER_LAST_UPLOAD_INFO_FILE_PATH) as json_file:
            multipartinfo = json.load(json_file)
            #Get the info
            uploadId=multipartinfo['uploadId']
            destination=multipartinfo['destination']
            sourceFile=multipartinfo['sourceFile']
            partsUpploaded=multipartinfo['partsUpploaded']
            isPrivate=multipartinfo['isPrivate']
            # get destination
            if "/" in destination and not isPrivate:
                key = destination.split('/', 1)[1]
                bucketName = destination.split('/')[0]
            else:
                key = destination
                bucketName = ""

            fileSize = os.stat(sourceFile).st_size
            print("Size "+ str(fileSize))
            #Set max to split to 5M
            max_size = 5 * 1024 * 1024 # Approach 1: Assign the size
            nbParts = math.ceil(fileSize/max_size)    #calculate nbParts
            finalPart = nbParts - len(partsUpploaded)
            startPart = len(partsUpploaded) + 1
            print("[INFO] We will have "+ str(nbParts)+" parts minus already uploaded parts. We have to push "+ str(finalPart) +" parts")
                        
            #Generate presigned urls
            url = "https://gravitee-gateway."+MAAP_ENV_TYPE.lower()+".esa-maap.org/s3/generateListPresignedUrls"
            params = {'bucketName': bucketName, 'objectKey': key, 'nbParts': nbParts, 'uploadId': uploadId,
                      'private': isPrivate}
            response = requests.get(url, params = params, headers = {'Authorization': 'Bearer '+token})
            if response.status_code == 401 or response.status_code == 403:
                print(response.content)
                sys.exit(1)

            stringList = response.text
            str1 = stringList.replace(']','').replace('[','')
            listPresignedUrl  = str1.replace('"','').split(",")    
            
            #Get the json uploaded list
            
            #we iterate over the data and repush
            with open(sourceFile, 'rb') as f:    
                i = 0
                while i < nbParts:
                    
                    file_data = f.read(max_size)
                    #We upload only if we have new nn uploaded part
                    if i >= len(partsUpploaded):
                        
                        print("Upload part "+ str(i+1))
                        headers={'Content-Length': str(max_size)}
                        try:
                            response= requests.put(listPresignedUrl[i], data=file_data, headers=headers)
                        except ConnectionError as e:
                            print("Error 401 ConnectionError")
                            print("Please retry later using resume option.")
                            sys.exit(1)

                        if response.status_code != 200:
                            print("Url : {}",listPresignedUrl[i])
                            print(response.text)
                            sys.exit(1)

                        etag = response.headers['ETag']  
                        partsUpploaded.append({'eTag': etag, 'partNumber': int(i+1)})
                        #We save also the multipart
                        #So we can resume it if upload failed
                        multipartinfo = {
                            'uploadId': uploadId,
                            'partsUpploaded': partsUpploaded,
                            'sourceFile': sourceFile,
                            'destination': destination,
                            'isPrivate': isPrivate
                        }
                        
                        #add the json in the file
                        with open(USER_LAST_UPLOAD_INFO_FILE_PATH, 'w') as outfile:
                            json.dump(multipartinfo, outfile)
                    
                    #We increase the data
                    i += 1
              
            #complete the multi part
            print("Complete multipart upload")
            url = "https://gravitee-gateway."+MAAP_ENV_TYPE.lower()+".esa-maap.org/s3/completeMultiPartUploadRequest"
            params={'bucketName': bucketName, 'objectKey': key, 'nbParts': nbParts, 'uploadId': uploadId, 'private': isPrivate}
            response = requests.get(url, data=str(partsUpploaded),  params = params, headers = {'Authorization': 'Bearer '+token})
            if response.status_code != 200:
                print(response.content)
                sys.exit(1)
            # delete the file of multipart info because upload was success
            print("Upload completed")
            os.remove(USER_LAST_UPLOAD_INFO_FILE_PATH) 
    
    else:
        print("[INFO] Please run upload. There are no upload to be resume")


###################
# Delete the data #
####################
@verify_token
def delete(destination, private):
    print("[INFO] Destination file is : ", destination)
    
    if destination:
        print("[INFO] Get an existing or fresh token")
        #Generate or get a token
        with open(USER_INFO_FILE_PATH) as json_file:
            userinfo = json.load(json_file)
            #Get the info
            token=userinfo['token']

        url = "https://gravitee-gateway."+MAAP_ENV_TYPE.lower()+".esa-maap.org/s3/"+destination
        list_object = requests.get(url+'?list=true', headers = {'Authorization': 'Bearer '+token}, allow_redirects=False)
        if list_object.status_code == 404:
            print(list_object.text)
            print('Object not found')
            sys.exit(0)
        elif list_object.status_code == 200:
            user_confirm = input("type y/Y and press enter to confirm deletion of the file : " + destination+" : ")
            if user_confirm not in ['y','Y']:
                print('Abort')
                sys.exit(0)


        #call the api to delete the data
        #Get the presigned url to delete the data
        url = "http://gravitee-gateway."+MAAP_ENV_TYPE.lower()+".esa-maap.org/s3/"+destination
        print(url)
        response = requests.delete(url,
                                   headers={'Authorization': 'Bearer ' + token},
                                   params={'private': str(private)},
                                   allow_redirects=False)
        if response.status_code == 401 or response.status_code == 403:
            print(response.content)
            sys.exit(1)

        location = response.headers['Location']
        
        #We have the 
        if location:
            #We delete the data using the location
            print("[INFO] we are about to delete")
            response = requests.delete(location)
            print(response)
        
    else:
        display_help()


###################
# download the data #
####################
@verify_token
def download(path, name, private: bool):
    print("[INFO] path file is : ", path)
    
    if path:
        print("[INFO] Get an existing or fresh token")
        #Generate or get a token
        with open(USER_INFO_FILE_PATH) as json_file:
            userinfo = json.load(json_file)
            #Get the info
            token=userinfo['token']
            
        #Get the presigned url to download the data
        url = "https://gravitee-gateway."+MAAP_ENV_TYPE.lower()+".esa-maap.org/s3/"+path
        response = requests.get(url, headers = {'Authorization': 'Bearer '+token}, params={'private': str(private)},
                                allow_redirects=False)
        if response.status_code == 401 or response.status_code == 403:
            print(response.content)
            sys.exit(1)
        location = response.headers['Location']
        
        #We have the 
        if location:
            #We download the data using the location
            print("[INFO] we are about to download the data")
            download_file(location, name)
            #response = requests.get(location)
            #open(name, 'wb').write(response.content)
            print("[INFO] Download finished")
        
    else:
        display_help()



##########################
# download file using url #
##########################
def download_file(url, name):
    #local_filename = url.split('/')[-1]
    # NOTE the stream=True parameter below
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(name, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192): 
                # If you have chunk encoded response uncomment if
                # and set chunk_size parameter to None.
                #if chunk: 
                f.write(chunk)
    return name


##########################
# list data in s3 folder #
##########################
@verify_token
def list(path, private: bool):
    print("[INFO]: Start finding data in path : "+path)
    
    if path:
        print("[INFO] Get an existing or fresh token")
        #Generate or get a token
        with open(USER_INFO_FILE_PATH) as json_file:
            userinfo = json.load(json_file)
            #Get the info
            token=userinfo['token']
            
        #call the api to delete the data
        #Get the presigned url to delete the data

        url = "https://gravitee-gateway."+MAAP_ENV_TYPE.lower()+".esa-maap.org/s3/"+path+"?list=true"

        response = requests.get(url, headers={'Authorization': 'Bearer ' + token}, params={'private': private},
                                allow_redirects=False)
        if response.status_code == 401:
            print(response.content)
            sys.exit(1)
        print("[INFO] Result list:")
        if(response.text):            
            print(response.text)
        else:
            print("[INFO] No data found")
    else:
        display_help()

# Store argument variable omitting the script name
argv = sys.argv[1:]

# Extract private option
private = False
if '-p' in argv:
    private = True
    argv.remove('-p')

if '--private' in argv:
    private = True
    argv.remove('--private')

# Initialize result variable
result=0
 
try:

    if len(argv) == 0:
        display_help()
    else:
        if argv[0] == 'resume':
            resume()
        elif argv[0] == 'refresh':
            refresh()
        elif argv[0] == 'upload':
            # Upload a data
            if len(argv) != 3:          
                display_help()
            else:
                upload(argv[1], argv[2], private)
        elif argv[0] == 'delete':
            # Delete a data
            if len(argv) != 2:
                display_help()
            else:
                delete(argv[1], private)
        elif argv[0] == 'token':
            # Delete a data
            if len(argv) != 3:
                display_help()
            else:
                get_token(argv[1], argv[2])
        elif argv[0] == 'login':
            # Delete a data
            if len(argv) != 3:
                display_help()
            else:
                login(argv[1], argv[2])    
        elif argv[0] == 'download':
            # Download a data
            if len(argv) != 3:
                display_help()
            else:
                download(argv[1], argv[2], private)
        elif argv[0] == 'list':
            # list a folder
            if len(argv) != 2:
                display_help()
            else:
                list(argv[1], private)
        elif argv[0] == 'help':
            display_help()
        else:  
            display_help()



except getopt.GetoptError:

  # Print the error message if the wrong option is provided
  print('The wrong option is provided. Please run -h')
 
  # Terminate the script
  sys.exit(2)

except Exception as e:
    print("exception: {}", e.__class__.__name__)
    print("Exception cause: {}", e.__cause__)
    sys.exit(1)
